/*
 * Access the LCD and related GPIOs on the Ben NanoNote.
 *
 * (c) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/cpm-jz4740.h>
#include <l4/devices/gpio-jz4740.h>
#include <l4/io/io.h>
#include <l4/re/env.h>
#include <l4/re/c/mem_alloc.h>
#include <l4/re/c/rm.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/cache.h>
#include <l4/util/util.h>
#include <l4/vbus/vbus.h>

#include <stdio.h>
#include <stdint.h>

#include "nanonote_gpm940b0.h"
#include "jzlcd.h"
#include "xburst_types.h"
#include "memory.h"

enum Jz4740_lcd_gpio
{
  Jz4740_lcd_gpio_spen = 21, /* serial command enable */
  Jz4740_lcd_gpio_spda = 22, /* serial command clock */
  Jz4740_lcd_gpio_spck = 23, /* serial command data */
};

/* Peripheral memory regions. */

static l4_addr_t cpm_virt_base = 0, cpm_virt_base_end = 0;
static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;
static l4_addr_t lcd_virt_base = 0, lcd_virt_base_end = 0;

/* Device abstractions. */

static void *gpio_port_c;
static void *cpm_device = 0;
static vidinfo_t *panel_info = 0;

/* Framebuffer and descriptor regions. */

static void *fb_vaddr = 0, *desc_vaddr = 0;
static l4_addr_t fb_paddr = 0, desc_paddr = 0;
static l4_size_t fb_size, desc_size;



// Write SPI values via the LCD GPIO pins.

static void spi_write_reg(uint8_t reg, uint8_t val)
{
  uint8_t no;
  uint16_t value;

  jz4740_gpio_set(gpio_port_c, Jz4740_lcd_gpio_spen, 1);
  jz4740_gpio_set(gpio_port_c, Jz4740_lcd_gpio_spck, 1);
  jz4740_gpio_set(gpio_port_c, Jz4740_lcd_gpio_spda, 0);
  jz4740_gpio_set(gpio_port_c, Jz4740_lcd_gpio_spen, 0);

  value = ((reg << 8) | (val & 0xFF));

  /* Clock data using the clock and data outputs. */

  for (no = 0; no < 16; no++)
  {
    jz4740_gpio_set(gpio_port_c, Jz4740_lcd_gpio_spck, 0);
    jz4740_gpio_set(gpio_port_c, Jz4740_lcd_gpio_spda, value & 0x8000 ? 1 : 0);
    jz4740_gpio_set(gpio_port_c, Jz4740_lcd_gpio_spck, 1);
    value = (value << 1);
  }

  jz4740_gpio_set(gpio_port_c, Jz4740_lcd_gpio_spen, 1);
}

// GPIO-based operations.

static void lcd_display_pin_init(void)
{ 
  Pin_slice mask = {.offset=0, .mask=(1 << Jz4740_lcd_gpio_spen) | (1 << Jz4740_lcd_gpio_spck) | (1 << Jz4740_lcd_gpio_spda)};
  Pin_slice slcd8_mask = {.offset=0, .mask=0x003c00ff};

  /* Configure SPI pins. */

  jz4740_gpio_multi_setup(gpio_port_c, &mask, L4VBUS_GPIO_SETUP_OUTPUT, 0);

  /* Configure SLCD8 pins. */

  jz4740_gpio_multi_config_pad(gpio_port_c, &slcd8_mask, Function_alt, 0);
}

static void lcd_display_on(void)
{ 
  spi_write_reg(0x05, 0x1e);  // GRB=0 (reset); PWM_DUTY=0b011 (70%, default); SHDB2=1, SHDB1=1 (power-related); STB=0 (standby)
  spi_write_reg(0x05, 0x5e);  // GRB=1 (normal operation); ...
  spi_write_reg(0x07, 0x8d);  // HBLK=141 (horizontal blanking period from start of hsync pulse to data start)
  spi_write_reg(0x13, 0x01);  // IN_SEL=1 (alignment mode)
  spi_write_reg(0x05, 0x5f);  // ...; STB=1 (not standby)
}

/* CPM operations. */

static void lcd_set_timing(vidinfo_t *vid)
{
  uint32_t pclk = jz4740_lcd_get_pixel_clock(vid);

  jz4740_cpm_stop_lcd(cpm_device);

  /*
  Original comment: LCDClock > 2.5*Pixclock
  However, the documentation indicates that a TFT panel needs a device clock
  1.5 times that of the pixel clock, and a STN panel needs a device clock 3
  times that of the pixel clock.
  */

  jz4740_cpm_set_lcd_frequencies(cpm_device, pclk, 3);
  jz4740_cpm_update_output_frequency(cpm_device);
  jz4740_cpm_start_lcd(cpm_device);

  l4_sleep(1); // 1ms == 1000us
}



static int setup_memory(void)
{
  l4re_ds_t fbmem, descmem;
  l4_size_t fb_size_out, desc_size_out;
  int result = 0;

  if (fb_vaddr)
    return 0;

  if (!panel_info)
    return 1;

  fb_size = jz4740_lcd_get_screen_size(panel_info);
  desc_size = jz4740_lcd_get_descriptors_size(panel_info);

  /* Obtain resource details describing the I/O memory. */

  if ((result = get_memory("jz4740-cpm", &cpm_virt_base, &cpm_virt_base_end)) < 0)
    return 1;

  if ((result = get_memory("jz4740-lcd", &lcd_virt_base, &lcd_virt_base_end)) < 0)
    return 1;

  if ((result = get_memory("jz4740-gpio", &gpio_virt_base, &gpio_virt_base_end)) < 0)
    return 1;

  /* Set the framebuffer up. */

  fbmem = l4re_util_cap_alloc();
  descmem = l4re_util_cap_alloc();

  if (l4_is_invalid_cap(fbmem) || l4_is_invalid_cap(descmem))
    return 1;

  /* Allocate memory for the framebuffer at 2**6 == 64 byte == 16 word alignment,
     also for the descriptors. */

  if (l4re_ma_alloc_align(fb_size, fbmem, L4RE_MA_CONTINUOUS | L4RE_MA_PINNED, 6) ||
      l4re_ma_alloc_align(desc_size, descmem, L4RE_MA_CONTINUOUS | L4RE_MA_PINNED, 6))
    return 1;

  /* Map the allocated memory, obtaining virtual addresses. */

  if (l4re_rm_attach(&fb_vaddr, fb_size,
                     L4RE_RM_SEARCH_ADDR | L4RE_RM_EAGER_MAP,
                     fbmem, 0, L4_PAGESHIFT) ||
      l4re_rm_attach(&desc_vaddr, desc_size,
                     L4RE_RM_SEARCH_ADDR | L4RE_RM_EAGER_MAP,
                     descmem, 0, L4_PAGESHIFT))
    return 1;

  fb_size_out = fb_size;
  desc_size_out = desc_size;

  if (l4re_ds_phys(fbmem, 0, &fb_paddr, &fb_size_out) ||
      l4re_ds_phys(descmem, 0, &desc_paddr, &desc_size_out))
    return 1;

  if ((fb_size_out != fb_size) || (desc_size_out != desc_size))
    return 1;

  cpm_device = jz4740_cpm_init(cpm_virt_base);
  gpio_port_c = jz4740_gpio_init(gpio_virt_base + 0x200, gpio_virt_base + 0x300, 32);

  return 0;
}

static void enable(void)
{
  if (setup_memory())
    return;

  jz4740_lcd_set_base(panel_info, (void *) lcd_virt_base);
  jz4740_lcd_disable(panel_info);

  // Initialise the LCD controller and structures.

  jz4740_lcd_ctrl_init(
    (struct jz_fb_dma_descriptor *) desc_vaddr,
    (struct jz_fb_dma_descriptor *) desc_paddr,
    fb_vaddr,
    (void *) fb_paddr,
    panel_info);

  // Initialise the LCD peripheral.

  jz4740_lcd_hw_init(panel_info);

  // Initialise the clocks for the LCD controller.

  lcd_set_timing(panel_info);

  // Switch the display on using GPIO operations.

  lcd_display_pin_init();
  lcd_display_on();

  // Finally, enable the peripheral.

  jz4740_lcd_enable(panel_info);
}



int main(void)
{
  l4_size_t *fb;
  l4_size_t i, mask, value, onpix, offpix;

  /* Configure the LCD. */

  panel_info = &nanonote_panel_info;
  enable();

  fb = (l4_size_t *) fb_vaddr;

  /* Try and show some values. */

  onpix = 0xffaaffaa; offpix = 0x11551155;
  mask = 0x80000000; value = 0x80000001;

  i = 0;

  while (i < fb_size / 4)
  {
    fb[i] = value & mask ? onpix : offpix;
    i++;

    if ((i % 10) == 0)
    {
      if (mask == 1) mask = 0x80000000;
      else mask >>= 1;

      onpix = (onpix >> 8) | ((onpix & 0xff) << 24);
      offpix = (offpix >> 8) | ((offpix & 0xff) << 24);

      if (i == 3200) value = jz4740_cpm_get_lcd_pixel_divider(cpm_device);
      else if (i == 6400) value = jz4740_cpm_get_lcd_pixel_frequency(cpm_device);
      else if (i == 9600) value = (l4_size_t) fb_vaddr;
      else if (i == 12800) value = fb_paddr;
      else if (i == 16000) value = ((struct jz_fb_dma_descriptor *) desc_vaddr)[0].fsadr;
      else if (i == 19200) value = REG32(lcd_virt_base + 0x00);
      else if (i == 22400) value = REG32(lcd_virt_base + 0x04);
      else if (i == 25600) value = REG32(lcd_virt_base + 0x30);
      else if (i == 28800) value = REG32(lcd_virt_base + 0x40);
    }
  }

  l4_cache_clean_data((long unsigned int) fb_vaddr, (long unsigned int) fb_vaddr + fb_size);

  while (1);

  return 0;
}
